import os
from enum import StrEnum

import pygame as pg
import numpy as np

from viztools.coordinate_system import CoordinateSystem
from viztools.drawable import Drawable


class OverlayPosition(StrEnum):
    TOP = 'top'
    LEFT = 'left'
    RIGHT = 'right'
    BOT = 'bot'
    RIGHTTOP = 'righttop'
    LEFTTOP = 'lefttop'
    LEFTBOT = 'leftbot'
    RIGHTBOT = 'rightbot'


class OverlayText(Drawable):
    def __init__(
            self,
            text: str,
            position: np.ndarray | OverlayPosition,
            absolute_position: bool = False,
            font_name: str = '',
            font_size: int | float = 16,
            color: np.ndarray | None = None,
            background_color: np.ndarray | None = None,
            border_color: np.ndarray | None = None,
            border_width: int = 2
    ):
        """
        Creates an overlay text.

        :param text: The text content to be displayed on the overlay.
        :param position: The position of the overlay in the form of a NumPy array [x, y] or an OverlayPosition object.
            If absolute_position, this is interpreted as absolute coordinates in the screen. Otherwise, it is
            interpreted as center point of the text in the coordinate system. If an OverlayPosition object is provided,
            it is always interpreted as absolute coordinates.
        :param absolute_position: If True, the position is interpreted as absolute coordinates in the screen.
        :param font_name: The font name to use for rendering the text. If not provided, defaults to a system font.
        :param font_size: The size of the font. If an integer is provided, it is interpreted as pixels. If a float is
            provided, it is interpreted as size in world coordinates.
        :param color: The color of the text as a NumPy array. If None, defaults to white. Does not support alpha.
        :param background_color: The background color of the text overlay as a NumPy array.
            If None, no background will be rendered.
        :param border_color: The color of the border around the text as a NumPy array. If
            None, no border will be rendered.
        :param border_width: The width of the border around the text, in pixels.
        """
        super().__init__()
        self.text = text
        self.position = position
        if isinstance(position, OverlayPosition):
            absolute_position = True
        self.absolute_position = absolute_position
        if font_name:
            if not os.path.isfile(font_name):
                font_name = pg.font.match_font(font_name)
        else:
            font_name = pg.font.get_default_font()
        self.font_name = font_name
        self.font_size = font_size
        self.color = color if color is not None else np.array([255, 255, 255], dtype=np.uint8)
        self.background_color = background_color
        self.border_color = border_color
        self.border_width = border_width

    def draw(self, screen: pg.Surface, coordinate_system: CoordinateSystem):
        text_lines = self.text.split('\n')
        font_size = self.font_size
        if isinstance(font_size, float):
            font_size = int(self.font_size * coordinate_system.zoom_factor)
        font = pg.font.Font(self.font_name, font_size)
        line_dimensions = [font.size(line) for line in text_lines]
        total_height = sum(d[1] for d in line_dimensions)
        max_width = max(d[0] for d in line_dimensions)

        combined_rect = pg.Rect(0, 0, max_width, total_height)
        padding = self.border_width * 2 if self.border_color is not None else 0

        if self.absolute_position:
            if isinstance(self.position, str):
                if self.position == OverlayPosition.TOP:
                    combined_rect.midtop = (screen.get_width() // 2, 0)
                elif self.position == OverlayPosition.LEFT:
                    combined_rect.midleft = (0, screen.get_height() // 2)
                elif self.position == OverlayPosition.RIGHT:
                    combined_rect.midright = (screen.get_width()-padding, screen.get_height() // 2)
                elif self.position == OverlayPosition.BOT:
                    combined_rect.midbottom = (screen.get_width() // 2, screen.get_height() - padding)
                elif self.position == OverlayPosition.RIGHTTOP:
                    combined_rect.topright = (screen.get_width() - padding, 0)
                elif self.position == OverlayPosition.LEFTTOP:
                    combined_rect.topleft = (0, 0)
                elif self.position == OverlayPosition.LEFTBOT:
                    combined_rect.bottomleft = (0, screen.get_height() - padding)
                elif self.position == OverlayPosition.RIGHTBOT:
                    combined_rect.bottomright = (screen.get_width() - padding, screen.get_height() - padding)
            else:
                combined_rect.topleft = self.position[:2]
        else:
            pos = coordinate_system.space_to_screen(self.position.reshape(2, 1)).reshape(2)
            combined_rect.center = (int(pos[0]), int(pos[1]))

        # Add padding for border
        combined_rect.width += padding
        combined_rect.height += padding

        # skip if out of screen
        if not combined_rect.colliderect(screen.get_rect()):
            return

        if self.background_color is not None:
            background = pg.Surface(combined_rect.size, pg.SRCALPHA)
            background.fill(self.background_color)
            screen.blit(background, combined_rect)

        if self.border_color is not None:
            border_surface = pg.Surface(screen.get_size(), pg.SRCALPHA)
            pg.draw.rect(border_surface, self.border_color, combined_rect, self.border_width)
            screen.blit(border_surface, (0, 0))

        current_y = combined_rect.y + (padding // 2)
        line_surfaces = [font.render(line, True, self.color) for line in text_lines]
        for surface in line_surfaces:
            line_rect = surface.get_rect()
            line_rect.centerx = combined_rect.centerx
            line_rect.y = current_y
            screen.blit(surface, line_rect)
            current_y += line_rect.height



import time
from typing import Iterable, Tuple, Dict, Optional

import pygame as pg
import numpy as np

from viztools.coordinate_system import CoordinateSystem
from viztools.drawable import Drawable
from viztools.drawable.draw_utils import normalize_color
from viztools.drawable.draw_utils.chunking import ChunkGrid


def get_viewport(coordinate_system, screen_size):
    return coordinate_system.screen_to_space_t(np.array([[0.0, 0.0], screen_size]))


class Points(Drawable):
    def __init__(
            self, points: np.ndarray, size: int | float | Iterable[int | float] = 3,
            color: np.ndarray | None = None
    ):
        """
        Drawable to display a set of points.
        :param points: A list of points with the shape [N, 2] where N is the number of points.
        :param size: The radius of the points. If set to an integer, this is the radius on the screen in pixels. If set
                     to a float, this is the radius on the screen in units of the coordinate system. If set to a list,
                     it contains the sizes for each point.
        :param color: The color of the points.
        values prevent unrendered chunks from being visible.
        """
        # points
        if not isinstance(points, np.ndarray):
            raise TypeError(f'points must be a numpy array, not {type(points)}.')
        if points.ndim != 2 or points.shape[1] != 2:
            raise ValueError(f'points must be numpy array with shape (N, 2), not {points.shape}.')
        n_points = len(points)
        self._points = points

        # size
        if isinstance(size, (int, float)):
            is_relative_size = isinstance(size, float)
            size = np.repeat(np.array([[size, float(is_relative_size)]], dtype=np.float32), n_points, axis=0)
        elif isinstance(size, np.ndarray):
            if size.shape != (n_points,):
                raise ValueError(f'size must be a numpy array with shape ({n_points},), not {size.shape}.')
            is_relative_size = np.full(n_points, np.issubdtype(size.dtype, np.floating), dtype=np.float32)
            size = np.stack([size.astype(np.float32), is_relative_size], axis=1)
        elif isinstance(size, list):
            if len(size) != n_points:
                raise ValueError(f'size must be a list of length {n_points}, not {len(size)}.')
            size = [[s, isinstance(s, float)] for s in size]
            size = np.array(size, dtype=np.float32)
        else:
            raise TypeError(f'size must be an integer, float or iterable, not {type(size)}.')
        self._size = size

        # colors
        if color is None:
            color = np.array([77, 178, 11])
        if isinstance(color, np.ndarray):
            if color.shape == (3,):
                color = np.array([*color, 255], dtype=np.float32)
            if color.shape == (4,):
                color = np.repeat(color.reshape(1, -1), n_points, axis=0).astype(np.float32)
        else:
            color = np.array(color, dtype=np.float32)
        if color.shape != (n_points, 4):
            raise ValueError(f'colors must be a numpy array with shape ({n_points}, 4), not {color.shape}.')
        self._colors = color

        self._surface_parameters = {}
        for surf_params in self._get_surf_params():
            self._surface_parameters[surf_params.tobytes()] = surf_params

        self.current_chunks: ChunkGrid = self._build_chunk_grid(100.0)
        self.last_zoom_factor = None

    def __len__(self):
        return len(self._points)

    def _build_chunk_grid(self, zoom_factor: float, chunk_size: int = 200) -> ChunkGrid:
        sizes = _get_world_sizes(self._size[:, 0], self._size[:, 1], zoom_factor)
        return ChunkGrid.from_points(self._points, sizes, chunk_size / zoom_factor)

    def _get_surf_params(self) -> np.ndarray:
        return np.concatenate([self._size, self._colors], axis=1)

    def _get_surf_param(self, index: int) -> np.ndarray:
        return np.concatenate([self._size[index, :], self._colors[index, :]], axis=0)

    def set_color(self, color: np.ndarray | Tuple[int, int, int, int], index: int):
        self._colors[index, :] = normalize_color(color)
        self._update_surf_params(index)

        # mark chunk to render new
        chunk_index = int(self.current_chunks.point_chunk_indices[index])
        self.current_chunks.set_status(chunk_index, 2)

    def _update_surf_params(self, index: int):
        surf_params = self._get_surf_param(index)
        self._surface_parameters[surf_params.tobytes()] = surf_params

    def set_size(self, size: int | float, index: int):
        self._size[index, 0] = size
        self._size[index, 1] = isinstance(size, float)
        self._update_surf_params(index)

    def _create_point_surfaces(self, zoom_factor: float) -> Dict[bytes, pg.Surface]:
        surfaces = {}
        for k, surf_params in self._surface_parameters.items():
            draw_size = _get_draw_size(surf_params[0], zoom_factor, bool(surf_params[1]))
            color = surf_params[2:]

            # old version with per pixel alpha
            point_surface = pg.Surface((draw_size * 2, draw_size * 2), pg.SRCALPHA)
            pg.draw.circle(point_surface, color, (draw_size, draw_size), draw_size)

            surfaces[k] = point_surface
        return surfaces

    def _get_draw_sizes(self, zoom_factor: float, sizes: Optional[np.ndarray] = None) -> np.ndarray:
        """
        Computes the draw sizes for the given sizes and coordinate system.
        :param zoom_factor: A float defining the scale factor for relative sizes.
        size[i] must be multiplied with zoom_factor.
        :return: numpy array of integers of shape [N,] where N is the number of sizes.
        """
        if sizes is None:
            sizes = self._size
        draw_sizes = sizes[:, 0].copy()
        is_relative_size = sizes[:, 1] > 0.5
        draw_sizes[is_relative_size] *= zoom_factor
        return np.maximum(draw_sizes.astype(int), 1)

    def update_chunks(self, coordinate_system: CoordinateSystem, screen_size: Tuple[int, int]) -> bool:
        point_surfaces = self._create_point_surfaces(coordinate_system.zoom_factor)

        if self.last_zoom_factor is None or self.last_zoom_factor != coordinate_system.zoom_factor:
            self.last_zoom_factor = coordinate_system.zoom_factor
            viewport = get_viewport(coordinate_system, screen_size)
            self.current_chunks.resize_chunks(coordinate_system.zoom_factor, viewport)

        start_time = time.perf_counter()
        while True:
            update_needed = self.render_next_chunk(coordinate_system, point_surfaces, screen_size)
            if not update_needed:
                return False
            if time.perf_counter() - start_time > 1 / 60:
                break
        return True

    def render_next_chunk(self, coordinate_system, point_surfaces, screen_size):
        viewport = get_viewport(coordinate_system, screen_size)
        update_index = self.current_chunks.get_next_update_chunk(viewport)
        if update_index is not None:
            sizes = _get_world_sizes(self._size[:, 0], self._size[:, 1], coordinate_system.zoom_factor)
            self.current_chunks.render_chunk(
                update_index, self._points, sizes, self._get_surf_params(), coordinate_system.zoom_factor,
                point_surfaces
            )
            return True
        return False

    def update(self, screen: pg.Surface, coordinate_system: CoordinateSystem) -> bool:
        return self.update_chunks(coordinate_system, screen.get_size())

    def draw(self, screen: pg.Surface, coordinate_system: CoordinateSystem):
        # draw points in chunks
        viewport = get_viewport(coordinate_system, screen.get_size())
        chunk_indices = self.current_chunks.get_in_viewport_chunk_indices(viewport)
        if self.current_chunks.get_pixel_approx(coordinate_system.zoom_factor) > 4000:
            # too many pixels: immediate mode

            # create blit surfaces
            surfaces = self._create_point_surfaces(coordinate_system.zoom_factor)

            # draw points in chunks
            for chunk_index in chunk_indices:
                chunk_index_tuple = self.current_chunks.chunk_index_tuple(chunk_index)
                point_indices = self.current_chunks.chunk_point_indices[chunk_index_tuple]
                screen_size = np.array(screen.get_size(), dtype=np.int32)

                # only consider points in chunk
                chunk_points = self._points[point_indices]
                chunk_sizes = self._size[point_indices]
                chunk_draw_sizes = self._get_draw_sizes(coordinate_system.zoom_factor, chunk_sizes)
                chunk_colors = self._colors[point_indices]
                chunk_surf_params = self._get_surf_params()[point_indices]

                # filter out points outside of screen
                screen_points = coordinate_system.space_to_screen_t(chunk_points)
                valid_positions = _get_valid_positions(screen_points, chunk_draw_sizes, screen_size)
                screen_points = screen_points[valid_positions]
                valid_sizes = chunk_draw_sizes[valid_positions]
                screen_points -= valid_sizes.reshape(-1, 1)
                valid_surf_params = chunk_surf_params[valid_positions]

                # draw
                for pos, surf_params in zip(screen_points, valid_surf_params):
                    surface = surfaces[surf_params.tobytes()]
                    screen.blit(surface, pos)
        else:
            for chunk_index in chunk_indices:
                chunk_x, chunk_y = self.current_chunks.chunk_index_tuple(chunk_index)
                if self.current_chunks.status[chunk_x, chunk_y] == 1:
                    self.current_chunks.resize_chunk((chunk_x, chunk_y), coordinate_system.zoom_factor)
                chunk_surface = self.current_chunks.get_surface((chunk_x, chunk_y))
                if chunk_surface is not None:
                    chunk_frame = self.current_chunks.get_chunk_frame((chunk_x, chunk_y))
                    left_top = np.array([[chunk_frame[0], chunk_frame[1]]])
                    left_top_screen = coordinate_system.space_to_screen_t(left_top)
                    left_top_screen = (int(left_top_screen[0, 0]), int(left_top_screen[0, 1]))
                    screen.blit(chunk_surface, left_top_screen)

    def clicked_points(self, event: pg.event.Event, coordinate_system: CoordinateSystem) -> np.ndarray:
        """
        Returns the indices of the points clicked by the mouse. Returns an empty array if no point was clicked.

        :param event: The event to check.
        :param coordinate_system: The coordinate system to use.
        """
        if event.type == pg.MOUSEBUTTONDOWN and event.button == 1:
            return self.hovered_points(np.array(event.pos, dtype=np.int32), coordinate_system)
        return np.array([])

    def hovered_points(self, mouse_pos: np.ndarray, coordinate_system: CoordinateSystem) -> np.ndarray:
        draw_sizes = self._get_draw_sizes(coordinate_system.zoom_factor)

        screen_pos = mouse_pos.reshape(1, 2)
        screen_points = coordinate_system.space_to_screen_t(self._points)
        distances = np.linalg.norm(screen_points - screen_pos, axis=1)
        return np.nonzero(distances < draw_sizes)[0]

    def closest_point(
            self, pos: np.ndarray, coordinate_system: CoordinateSystem, dist_to_center: bool = False
    ) -> Tuple[int, float]:
        """
        Finds the closest point to the given position.

        This function calculates the closest point to a specified 2D position on
        the screen.

        :param pos: The 2D position in screen coordinates to calculate the distance from.
        :param coordinate_system: The coordinate system used for transforming space
                                  coordinates to screen coordinates.
        :param dist_to_center: If False, the distance is calculated as the distance between the edge of point and <pos>.
            If True, the distance is calculated as the distance between the center of the point and <pos>.
        :return: A tuple containing the index of the closest point and the distance
                 to that closest point.
        :rtype: Tuple[int, float]
        """
        screen_pos = pos.reshape(1, 2)
        screen_points = coordinate_system.space_to_screen_t(self._points)
        distances = np.linalg.norm(screen_points - screen_pos, axis=1)
        if not dist_to_center:
            distances -= self._get_draw_sizes(coordinate_system.zoom_factor)
        closest_index = np.argmin(distances)
        return int(closest_index), max(float(distances[closest_index]), 0.0)


def _get_draw_size(
        size: float, zoom_factor: float, is_relative_size: bool
) -> int:
    if is_relative_size:
        size = max(int(size * zoom_factor), 1)
    return int(size)


def _get_world_sizes(sizes: np.ndarray, is_relative: np.ndarray, zoom_factor: float) -> np.ndarray:
    """
    Converts sizes from screen or world sizes to world sizes.
    :param sizes: Float array with shape [N,] where N is the number of sizes.
    :param is_relative: Float array with shape [N,] where N is the number of sizes. 1.0 means relative size, 0.0 means
    absolute size.
    :param zoom_factor: The zoom factor. world_size = draw_size / zoom_factor
    :return: For each index i, returns the world size for size[i]
    """
    is_absolute = is_relative < 0.5
    world_sizes = sizes.copy()
    world_sizes[is_absolute] /= zoom_factor
    return world_sizes


def _get_valid_positions(screen_points: np.ndarray, draw_sizes: np.ndarray, screen_size: np.ndarray) -> np.ndarray:
    return np.where(np.logical_and(
        np.logical_and(screen_points[:, 0] > -draw_sizes, (screen_points[:, 0] < screen_size[0] + draw_sizes)),
        np.logical_and(screen_points[:, 1] > -draw_sizes, (screen_points[:, 1] < screen_size[1] + draw_sizes))
    ))[0]



import numpy as np
import pygame as pg

from viztools.coordinate_system import CoordinateSystem


class CoordinateSystemController:
    def __init__(self, coordinate_system: CoordinateSystem, drag_mouse_button: int = 2):
        self.coordinate_system = coordinate_system
        self.dragging: bool = False
        self.mouse_position = np.zeros(2, dtype=int)
        self.drag_mouse_button = drag_mouse_button

    def handle_event(self, event: pg.event.Event) -> bool:
        """

        :param event:
        :return:
        """
        render_needed = False
        if event.type == pg.MOUSEBUTTONDOWN:
            if event.button == self.drag_mouse_button:
                self.dragging = True
                render_needed = True
        elif event.type == pg.MOUSEBUTTONUP:
            if event.button == self.drag_mouse_button:
                self.dragging = False
                render_needed = True
        elif event.type == pg.MOUSEMOTION:
            self.mouse_position = np.array(event.pos, dtype=np.int32)
            if self.dragging:
                self.coordinate_system.translate(np.array(event.rel, dtype=np.int32))
                render_needed = True
        elif event.type == pg.MOUSEWHEEL:
            if event.y < 0:
                self.coordinate_system.zoom_out(focus_point=self.mouse_position)
                render_needed = True
            else:
                self.coordinate_system.zoom_in(focus_point=self.mouse_position)
                render_needed = True
        return render_needed



import warnings
from typing import Self, Tuple, Optional, Dict

import numpy as np
import pygame as pg


class ChunkGrid:
    def __init__(
            self, surfaces: np.ndarray, points: np.ndarray, sizes: np.ndarray, point_chunk_indices: np.ndarray,
            chunk_point_indices: np.ndarray, left_bot: np.ndarray, chunk_size: float, chunk_frames: np.ndarray,
    ):
        """
        Creates a new ChunkGrid object. This groups the given points into a grid of chunks with size (w, h).
        Each chunk consists of a pg.Surface, a rect in world coordinates defining which points are in the chunk and
        points-indices of the points in the chunk.

        :param surfaces: A numpy array with shape (w, h) of type pg.Surface.
        :param points: The center positions of points in the coordinate system with shape (n, 2).
        :param sizes: The sizes of the points in the coordinate system with shape n.
        :param point_chunk_indices: A numpy array with shape n of type int. Accessing point_indices[i] gives the
        chunk_index of the ith point. Left bottom chunk has index 0, the next chunk to the top has index 1.
        The top right chunk as index h*w-1.
        :param chunk_point_indices: A numpy array with shape (w, h).
        Accessing chunk_point_indices[x, y] gives a numpy array of indices for points in the chunk at (x, y).
        :param left_bot: The world coordinates of the bottom left corner of the bottom left chunk as a numpy array with
        shape 2 of type float.
        :param chunk_size: The size of the chunks in world coordinates.
        :param chunk_frames: A numpy array with shape (w, h, 5). Each entry[x, y] contains the
        (init, left, top, right, bottom) world coordinates of the surrounding frame. If init is 0.0, the frame is not
        initialized. The surrounding frame is the area, in which all contained points can be fully rendered.
        If no points are contained in the chunk, the frame should be zeros.
        """
        self.surfaces = surfaces
        self.points = points
        self.sizes = sizes
        self.point_chunk_indices = point_chunk_indices
        self.chunk_point_indices = chunk_point_indices
        self.left_bot = left_bot
        self.chunk_size = chunk_size
        # meaning of status:
        # - 0: not rendered
        # - 1: has to rescale
        # - 2: has to update
        # - 3: ok
        self.status = np.zeros(surfaces.shape, dtype=np.int32)
        self.chunk_frames = chunk_frames

    def shape(self) -> Tuple[int, int]:
        """
        The number of chunks in the grid as tuple (w, h).
        """
        return self.surfaces.shape

    @classmethod
    def from_points(cls, points: np.ndarray, sizes: np.ndarray, chunk_size: float) -> Self:
        """
        Create a grid of chunks from the given points. Each point is in exactly one chunk.
        Chunks are created in scanline order.

        :param points: The center positions of points in the coordinate system with shape (n, 2).
        Accessing points[i] gives the (x, y) position of point i.
        :param sizes: The sizes of the points in the coordinate system with shape n.
        :param chunk_size: The maximum size of a chunk in world size.
        :return: A ChunkGrid object.
        """
        if points.ndim != 2 or points.shape[1] != 2:
            raise ValueError(f'points must be a numpy array with shape (n, 2), not {points.shape}.')
        if not np.issubdtype(points.dtype, np.floating):
            raise ValueError(f'points must be a numpy array with floating point values, not {points.dtype}.')

        most_left_bot = np.min(points, axis=0)
        most_right_top = np.max(points, axis=0)

        world_size = most_right_top - most_left_bot
        chunks_shape = np.trunc(world_size / chunk_size).astype(np.int32) + 1

        surfaces = np.full(chunks_shape, None, dtype=object)  # numpy array of pg.Surface

        points_rel_to_bot_left = points - most_left_bot.reshape(1, 2)
        chunk_indices_per_axis = np.trunc(points_rel_to_bot_left / chunk_size)
        for dim in range(2):
            chunk_indices_per_axis[:, dim] = np.clip(chunk_indices_per_axis[:, dim], 0, chunks_shape[dim]-1)

        # index = x * h + y
        point_chunk_indices: np.ndarray = chunk_indices_per_axis[:, 1] + chunk_indices_per_axis[:, 0] * chunks_shape[1]

        chunk_point_indices = np.full(chunks_shape, None, dtype=object)

        # building chunk frames
        chunk_frames = np.zeros((*chunks_shape, 5))

        return ChunkGrid(
            surfaces, points, sizes, point_chunk_indices, chunk_point_indices, most_left_bot, chunk_size, chunk_frames
        )
    
    def get_in_viewport_chunk_indices(self, viewport: np.ndarray) -> np.ndarray:
        """
        Get chunk indices in the viewport.

        :param viewport: Numpy array of shape (2, 2) with the viewport coordinates in world coordinates.
        Accessing viewport[0] gives the top left corner of the viewport in world coordinates. Accessing viewport[1]
        gives the bottom right corner of the viewport in world coordinates.
        """
        # Convert viewport coordinates to chunk indices
        rel_viewport = (viewport - self.left_bot.reshape(1, 2)) / self.chunk_size
        left = np.trunc(rel_viewport[0, 0]).astype(np.int32)
        top = np.ceil(rel_viewport[0, 1]).astype(np.int32) - 1
        right = np.ceil(rel_viewport[1, 0]).astype(np.int32) - 1
        bot = np.trunc(rel_viewport[1, 1]).astype(np.int32)

        left_right = np.clip([left, right], 0, self.shape()[0]-1)
        bot_top = np.clip([bot, top], 0, self.shape()[1]-1)

        # Create grid of all chunk positions between top-left and bottom-right
        x = np.arange(left_right[0], left_right[1] + 1)
        y = np.arange(bot_top[0], bot_top[1] + 1)
        x_axis, y_axis = np.meshgrid(x, y)

        # Convert to linear indices (x * h + y)
        return x_axis.flatten() * self.shape()[1] + y_axis.flatten()

    def get_next_update_chunk(self, viewport: np.ndarray) -> Optional[int]:
        """
        Calculates the chunk index of the next chunk to draw.
        
        :param viewport: Numpy array of shape (2, 2) with the viewport coordinates in world coordinates.
        Accessing viewport[0] gives the left top corner of the viewport in world coordinates. Accessing viewport[1]
        gives the right bottom corner of the viewport in world coordinates.
        """
        update_chunk = self._get_next_update_chunk_impl(viewport)
        if update_chunk is not None:
            return update_chunk
        else:
            width = abs(viewport[1, 0] - viewport[0, 0])
            height = abs(viewport[1, 1] - viewport[0, 1])
            viewport_extension = 1.0
            extended_viewport = np.array([
                [viewport[0, 0] - width * viewport_extension, viewport[0, 1] + height * viewport_extension],
                [viewport[1, 0] + width * viewport_extension, viewport[1, 1] - height * viewport_extension]
            ])
            return self._get_next_update_chunk_impl(extended_viewport)

    def _get_next_update_chunk_impl(self, viewport: np.ndarray) -> Optional[int]:
        """
        Calculates the chunk index of the next chunk to draw.

        :param viewport: Numpy array of shape (2, 2) with the viewport coordinates in world coordinates.
        Accessing viewport[0] gives the top left corner of the viewport in world coordinates. Accessing viewport[1]
        gives the bottom right corner of the viewport in world coordinates.
        """
        chunk_indices = self.get_in_viewport_chunk_indices(viewport)
        chunk_status = self.status.flat[chunk_indices]
        most_needed_index = np.argmin(chunk_status)
        if chunk_status[most_needed_index] == 3:
            return None
        return int(chunk_indices[most_needed_index])

    def set_status(self, chunk_index: int, status: int):
        self.status[self.chunk_index_tuple(chunk_index)] = status

    def chunk_index_tuple(self, chunk_index: int) -> Tuple[int, int]:
        s = self.shape()
        return chunk_index // s[1], chunk_index % s[1]

    def chunk_frame_size(self, chunk_index: int) -> Tuple[float, float]:
        frame = self.chunk_frames[self.chunk_index_tuple(chunk_index)]
        if frame[0] > 0.5:
            return abs(float(frame[2] - frame[0])), abs(float(frame[3] - frame[1]))
        warnings.warn(f'using approximation for frame size for chunk {chunk_index}')
        return self.chunk_size, self.chunk_size

    def get_chunk_point_indices(self, chunk_index_tuple: Tuple[int, int]) -> np.ndarray:
        chunk_x, chunk_y = chunk_index_tuple
        if self.chunk_point_indices[chunk_index_tuple] is None:
            chunk_index: int = chunk_index_tuple[0] * self.shape()[1] + chunk_index_tuple[1]
            cur_chunk = np.nonzero(np.equal(self.point_chunk_indices, chunk_index))[0]
            self.chunk_point_indices[chunk_x, chunk_y] = cur_chunk
        return self.chunk_point_indices[chunk_index_tuple]

    def get_chunk_frame(self, chunk_index_tuple: Tuple[int, int]) -> np.ndarray:
        frame = self.chunk_frames[chunk_index_tuple]
        if frame[0] < 0.5:
            point_indices = self.get_chunk_point_indices(chunk_index_tuple)
            if point_indices.shape[0] != 0:
                point_positions = self.points[point_indices]
                point_sizes = self.sizes[point_indices].reshape(-1, 1)

                left_bot = np.min(point_positions - point_sizes, axis=0)
                right_top = np.max(point_positions + point_sizes, axis=0)
                self.chunk_frames[chunk_index_tuple] = (1.0, left_bot[0], right_top[1], right_top[0], left_bot[1])
        return self.chunk_frames[chunk_index_tuple][1:]

    def invalidate_chunks(self):
        shape = self.shape()
        self.surfaces = np.full(shape, None, dtype=object)  # numpy array of pg.Surface
        self.status = np.zeros(shape, dtype=np.int32)

    def get_surface(self, chunk_index_tuple: Tuple[int, int]) -> Optional[pg.Surface]:
        # noinspection PyTypeChecker
        return self.surfaces[chunk_index_tuple]

    def resize_chunks(self, zoom_factor: float, viewport: np.ndarray):
        self.status[:] = 1  # everything has to be rescaled
        for chunk_index in self.get_in_viewport_chunk_indices(viewport):
            chunk_index = self.chunk_index_tuple(chunk_index)
            self.resize_chunk(chunk_index, zoom_factor)

    def resize_chunk(self, chunk_index_tuple: Tuple[int, int], zoom_factor: float):
        chunk_x, chunk_y = chunk_index_tuple
        _frame, render_size = self._get_render_frame_size(chunk_index_tuple, zoom_factor)
        # don't render chunks with too many pixels
        if np.prod(render_size) > 4000 ** 2:
            return
        current_surface = self.get_surface((chunk_x, chunk_y))
        # we can't scale if the surface has not been rendered yet
        if current_surface is None:
            return
        new_surface = pg.transform.scale(current_surface, render_size)
        self.surfaces[chunk_x, chunk_y] = new_surface
        self.status[chunk_x, chunk_y] = 2

    def render_chunk(
            self, chunk_index: int, points: np.ndarray, sizes: np.ndarray, surf_params: np.ndarray,
            zoom_factor: float, point_surfaces: Dict[bytes, pg.Surface]
    ):
        """
        Creates a surface for the given chunk
        :param chunk_index:
        """
        chunk_index = self.chunk_index_tuple(chunk_index)

        frame, render_size = self._get_render_frame_size(chunk_index, zoom_factor)
        # don't render chunks with too many pixels
        if np.prod(render_size) > 4000 ** 2:
            return
        surface = pg.Surface(render_size, pg.SRCALPHA)

        point_indices = self.chunk_point_indices[chunk_index]
        points = points[point_indices]
        sizes = sizes[point_indices]
        surf_params = surf_params[point_indices]

        left_bot = np.array([frame[0], frame[3]])
        render_positions = points - left_bot.reshape(1, 2)
        render_positions[:, 0] -= sizes
        render_positions[:, 1] += sizes
        render_positions *= zoom_factor
        render_positions[:, 1] = render_size[1] - render_positions[:, 1]  # flip y-axis
        render_positions = render_positions.round().astype(int)

        for pos, surf_params in zip(render_positions, surf_params):
            point_surface = point_surfaces[surf_params.tobytes()]
            surface.blit(point_surface, pos)

        self.status[chunk_index] = 3
        self.surfaces[chunk_index] = surface

    def get_pixel_approx(self, zoom_factor) -> int:
        return int(self.chunk_size * zoom_factor)

    def _get_render_frame_size(self, chunk_index, zoom_factor):
        frame = self.get_chunk_frame(chunk_index)
        frame_size = abs(float(frame[2] - frame[0])), abs(float(frame[3] - frame[1]))
        render_size = tuple(int(round(s * zoom_factor)) for s in frame_size)
        return frame, render_size



from typing import Tuple

import numpy as np


Color = np.ndarray | Tuple[int, int, int, int] | Tuple[int, int, int]


def normalize_color(color: Color) -> np.ndarray:
    if len(color) == 3:
        return np.array([*color, 255], dtype=np.float32)
    if len(color) != 4:
        raise ValueError(f'color must be of length 3 or 4, not {len(color)}.')
    return np.array(color, dtype=np.float32)



from abc import ABC, abstractmethod

import pygame as pg

from viztools.coordinate_system import CoordinateSystem


class Drawable(ABC):
    @abstractmethod
    def draw(self, screen: pg.Surface, coordinate_system: CoordinateSystem):
        pass

    def update(self, screen: pg.Surface, coordinate_system: CoordinateSystem) -> bool:
        return False
